import numpy as np
import cv2
import zebrazoom.code.popUpAlgoFollow as popUpAlgoFollow
from zebrazoom.code.preprocessImage import preprocessImage, preprocessBackgroundImage

def getBackground(videoPath, hyperparameters):

  cap   = cv2.VideoCapture(videoPath)
  max_l = int(cap.get(7))

  backCalculationStep = hyperparameters["backCalculationStep"]
  if ("firstFrame" in hyperparameters) and (hyperparameters["backgroundExtractionForceUseAllVideoFrames"] == 0):
    firstFrame = hyperparameters["firstFrame"]
  else:
    firstFrame = 1
  if ("lastFrame" in hyperparameters) and (hyperparameters["backgroundExtractionForceUseAllVideoFrames"] == 0):
    lastFrame  = hyperparameters["lastFrame"]
  else:
    lastFrame  = max_l - 10
  debugExtractBack    = hyperparameters["debugExtractBack"]
  
  if backCalculationStep == -1:
    backCalculationStep = int((lastFrame - firstFrame) / hyperparameters["nbImagesForBackgroundCalculation"])
    if backCalculationStep <= 1:
      backCalculationStep = 1
  
  ret, back = cap.read()
  if ret and hyperparameters["invertBlackWhiteOnImages"]:
    back = 255 - back
  back = cv2.cvtColor(back, cv2.COLOR_BGR2GRAY)
  if hyperparameters["backgroundExtractionWithOnlyTwoFrames"] == 0:
    for k in range(firstFrame,lastFrame):
      if (k % backCalculationStep == 0):
        cap.set(1, k)
        ret, frame = cap.read()
        if ret and hyperparameters["invertBlackWhiteOnImages"]:
          frame = 255 - frame
        if debugExtractBack:
          print(k)
        if ret:
          frame = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
          if hyperparameters["extractBackWhiteBackground"]:
            back = cv2.max(frame, back)
          else:
            back = cv2.min(frame, back)
        else:
          print("couldn't use the frame", k, "for the background extraction")
  else:
    maxDiff    = 0
    indMaxDiff = firstFrame
    for k in range(firstFrame,lastFrame):
      if (k % backCalculationStep == 0):
        cap.set(1, k)
        ret, frame = cap.read()
        if ret:
          if hyperparameters["invertBlackWhiteOnImages"]:
            frame = 255 - frame
          frame = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
          diff  = np.sum(np.abs(frame - back))
          if diff > maxDiff:
            maxDiff    = diff
            indMaxDiff = k
    cap.set(1, indMaxDiff)
    ret, frame = cap.read()
    if ret:
      if hyperparameters["invertBlackWhiteOnImages"]:
        frame = 255 - frame
      frame = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
      if hyperparameters["extractBackWhiteBackground"]:
        back = cv2.max(frame, back)
      else:
        back = cv2.min(frame, back)
    else:
      print("couldn't use the frame", k, "for the background extraction")
  
  if hyperparameters["backgroundPreProcessMethod"]:
    back = preprocessBackgroundImage(back, hyperparameters)
  
  if hyperparameters["checkThatMovementOccurInVideo"]:
    cap.set(1, firstFrame)
    ret, frame = cap.read()
    if ret:
      if hyperparameters["invertBlackWhiteOnImages"]:
        frame = 255 - frame
      frame = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
      if hyperparameters["imagePreProcessMethod"]:
        frame = preprocessBackgroundImage(frame, hyperparameters)
      if type(frame[0][0]) == np.ndarray:
        frame = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
      putToWhite = (frame.astype('int32') >= (back.astype('int32')-hyperparameters["minPixelDiffForBackExtract"]))
      frame[putToWhite] = 255
    firstImage = frame
    maxDiff    = 0
    indMaxDiff = firstFrame
    for k in range(firstFrame,lastFrame):
      if (k % backCalculationStep == 0):
        cap.set(1, k)
        ret, frame = cap.read()
        if ret:
          if hyperparameters["invertBlackWhiteOnImages"]:
            frame = 255 - frame
          frame = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
          if hyperparameters["imagePreProcessMethod"]:
            frame = preprocessBackgroundImage(frame, hyperparameters)
          if type(frame[0][0]) == np.ndarray:
            frame = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
          putToWhite = (frame.astype('int32') >= (back.astype('int32')-hyperparameters["minPixelDiffForBackExtract"]))
          frame[putToWhite] = 255
          if hyperparameters["checkThatMovementOccurInVideoMedianFilterWindow"]:
            diff  = np.sum(cv2.medianBlur(np.abs(frame - firstImage), hyperparameters["checkThatMovementOccurInVideoMedianFilterWindow"]))
          else:
            diff  = np.sum(np.abs(frame - firstImage))
          if diff > maxDiff:
            maxDiff    = diff
            indMaxDiff = k
    print("checkThatMovementOccurInVideo: max difference is:", maxDiff)
    if maxDiff < hyperparameters["checkThatMovementOccurInVideo"]:
      back[:, :] = 0 # TODO: tracking should NOT RUN after background is set to 0 as it is here
  
  if (debugExtractBack):
      
    cv2.imshow('Background extracted', back)
    if hyperparameters["exitAfterBackgroundExtraction"]:
      cv2.waitKey(3000)
    else:
      cv2.waitKey(0)
    cv2.destroyAllWindows()
    
  cap.release()
  
  print("Background Extracted")
  if hyperparameters["popUpAlgoFollow"]:
    popUpAlgoFollow.prepend("Background Extracted")
  
  return back



import numpy as np
import cv2
import math

def postProcessMultipleTrajectories(trackingHeadingAllAnimals, trackingHeadTailAllAnimals, trackingEyesAllAnimals, hyperparameters):
  
  maxDistanceAuthorized = hyperparameters["postProcessMaxDistanceAuthorized"]
  maxDisapearanceFrames = hyperparameters["postProcessMaxDisapearanceFrames"]
  
  currentlyZero  = False
  zeroFrameStart = 0
  for animalId in range(0, len(trackingHeadTailAllAnimals)):
    for frameNumber in range(0, len(trackingHeadTailAllAnimals[animalId])):
      xHead = trackingHeadTailAllAnimals[animalId][frameNumber][0][0]
      yHead = trackingHeadTailAllAnimals[animalId][frameNumber][0][1]
      if frameNumber != 0:
        xHeadPrev = trackingHeadTailAllAnimals[animalId][frameNumber-1][0][0]
        yHeadPrev = trackingHeadTailAllAnimals[animalId][frameNumber-1][0][1]
      else:
        xHeadPrev = trackingHeadTailAllAnimals[animalId][frameNumber][0][0]
        yHeadPrev = trackingHeadTailAllAnimals[animalId][frameNumber][0][1]
      if (xHead == 0 and yHead == 0) or (math.sqrt((xHead - xHeadPrev)**2 + (yHead - yHeadPrev)**2) > maxDistanceAuthorized):
        if not(currentlyZero):
          zeroFrameStart = frameNumber
        print("currentlyZero at True: animalId:", animalId, "; frameNumber:", frameNumber)
        currentlyZero = True
      else:
        if currentlyZero:
          if zeroFrameStart >= 1:
            xHeadStart = trackingHeadTailAllAnimals[animalId][zeroFrameStart-1][0][0]
            yHeadStart = trackingHeadTailAllAnimals[animalId][zeroFrameStart-1][0][1]
          else:
            xHeadStart = trackingHeadTailAllAnimals[animalId][frameNumber][0][0]
            yHeadStart = trackingHeadTailAllAnimals[animalId][frameNumber][0][1]
          xHeadEnd = trackingHeadTailAllAnimals[animalId][frameNumber][0][0]
          yHeadEnd = trackingHeadTailAllAnimals[animalId][frameNumber][0][1]
          if (math.sqrt((xHeadEnd - xHeadStart)**2 + (yHeadEnd - yHeadStart)**2) < maxDistanceAuthorized) or (frameNumber - zeroFrameStart > maxDisapearanceFrames):
            currentlyZero = False
            if (math.sqrt((xHeadEnd - xHeadStart)**2 + (yHeadEnd - yHeadStart)**2) < maxDistanceAuthorized):
              xStep = (xHeadEnd - xHeadStart) / (frameNumber - zeroFrameStart)
              yStep = (yHeadEnd - yHeadStart) / (frameNumber - zeroFrameStart)
              for frameAtZeroToChange in range(zeroFrameStart, frameNumber):
                trackingHeadTailAllAnimals[animalId][frameAtZeroToChange][0][0] = xHeadStart + xStep * (frameAtZeroToChange - zeroFrameStart)
                trackingHeadTailAllAnimals[animalId][frameAtZeroToChange][0][1] = yHeadStart + yStep * (frameAtZeroToChange - zeroFrameStart)    
            else:
              xStep = (xHeadEnd - xHeadStart) / (frameNumber - zeroFrameStart)
              yStep = (yHeadEnd - yHeadStart) / (frameNumber - zeroFrameStart)
              for frameAtZeroToChange in range(zeroFrameStart, frameNumber):
                trackingHeadTailAllAnimals[animalId][frameAtZeroToChange][0][0] = xHeadStart
                trackingHeadTailAllAnimals[animalId][frameAtZeroToChange][0][1] = yHeadStart
  
  return [trackingHeadingAllAnimals, trackingHeadTailAllAnimals, trackingEyesAllAnimals]



import cv2
import os
import shutil
import math
import matplotlib.pyplot as plt
import numpy as np
import time
import pandas as pd
import pickle
import scipy as sp

def perBoutOutput(superStruct, hyperparameters, videoName):
  
  # Creation of the sub-folder "perBoutOutput" 
  outputPath = os.path.join(os.path.join(hyperparameters["outputFolder"], videoName), "perBoutOutput")
  if os.path.exists(outputPath):
    shutil.rmtree(outputPath)
  while True:
    try:
      os.mkdir(outputPath)
      break
    except OSError as e:
      print("waiting inside except")
      time.sleep(0.1)
    else:
      print("waiting")
      time.sleep(0.1)
      
  # opening the video previously created in createValidationVideo as an input video stream
  cap = cv2.VideoCapture(os.path.join(os.path.join(hyperparameters["outputFolder"], videoName), hyperparameters["videoName"] + '.avi'))
  
  if (cap.isOpened()== False): 
    print("Error opening video stream or file")
  
  # Video frame start and end
  firstFrame = hyperparameters["firstFrame"]
  lastFrame  = hyperparameters["lastFrame"]
  nbFrames   = lastFrame - firstFrame + 1
  perBoutOutputVideoStartStopFrameMargin = hyperparameters["perBoutOutputVideoStartStopFrameMargin"]
  
  # Getting well positions
  infoWells = []
  for i in range(0, len(superStruct["wellPositions"])):
    x = superStruct["wellPositions"][i]["topLeftX"]
    y = superStruct["wellPositions"][i]["topLeftY"]
    lengthX = superStruct["wellPositions"][i]["lengthX"]
    lengthY = superStruct["wellPositions"][i]["lengthY"]
    infoWells.append([x, y, lengthX, lengthY])
  
  # Going through each well, each fish and each bout
  for i in range(0, len(superStruct["wellPoissMouv"])):
    for j in range(0, len(superStruct["wellPoissMouv"][i])):
      for k in range(0, len(superStruct["wellPoissMouv"][i][j])):
        
        # Creation of the curvature graph for bout k
        TailX_VideoReferential = superStruct["wellPoissMouv"][i][j][k]["TailX_VideoReferential"]
        TailY_VideoReferential = superStruct["wellPoissMouv"][i][j][k]["TailY_VideoReferential"]
        
        curvature = []
        for l in range(0, len(TailX_VideoReferential)):
          tailX = TailX_VideoReferential[l]
          tailY = TailY_VideoReferential[l]

          ydiff  = np.diff(tailY)
          ydiff2 = np.diff(ydiff)
          xdiff  = np.diff(tailX)
          xdiff2 = np.diff(xdiff)
          curv = xdiff2
          l = len(curv)
          av = 0
          for ii in range(0, l-1):
            num = xdiff[ii] * ydiff2[ii] - ydiff[ii] * xdiff2[ii]
            den = (xdiff[ii]**2 + ydiff[ii]**2)**1.5
            curv[ii] = num / den
          
          curv = curv[hyperparameters["nbPointsToIgnoreAtCurvatureBeginning"]: l-1-hyperparameters["nbPointsToIgnoreAtCurvatureEnd"]]
          curvature.append(curv)
          
        if False: # Old version: 1d median filter: TO REMOVE MOST LIKELY
        
          rolling_window = hyperparameters["curvatureMedianFilterSmoothingWindow"]
          if rolling_window:
            curvature2 = []
            curvatureTransposed = np.transpose(curvature)
            for transposedCurv in curvatureTransposed:
              shift = int(-rolling_window / 2)
              smoothedTransposedCurv = np.array(pd.Series(transposedCurv).rolling(rolling_window).median())
              smoothedTransposedCurv = np.roll(smoothedTransposedCurv, shift)
              for ii in range(0, rolling_window):
                smoothedTransposedCurv[ii] = transposedCurv[ii]
              for ii in range(len(smoothedTransposedCurv)-rolling_window,len(smoothedTransposedCurv)):
                smoothedTransposedCurv[ii] = transposedCurv[ii] 
              curvature2.append(smoothedTransposedCurv)  
            curvature = np.transpose(curvature2)
          else:
            curvature = np.array(curvature)
            
        else: # New version: 2d median filter
        
          rolling_window = hyperparameters["curvatureMedianFilterSmoothingWindow"]
          if rolling_window:
            curvature = sp.signal.medfilt2d(curvature, rolling_window)
          else:
            curvature = np.array(curvature)
        
        fig = plt.figure(1)
        plt.pcolor(curvature)
        
        ax = fig.axes
        ax[0].set_xlabel('Rostral to Caudal')
        ax[0].set_ylabel('Frame number')
        plt.colorbar()
        plt.savefig(os.path.join(outputPath, hyperparameters["videoName"] + "_curvature_bout" + str(i) + '_' + str(j) + '_' + str(k) + '.png'))
        plt.close(1)
        
        curvatureMatrixFile = open(os.path.join(outputPath, hyperparameters["videoName"] + "_curvatureData" + str(i) + '_' + str(j) + '_' + str(k) + '.txt'), 'wb')
        pickle.dump(curvature, curvatureMatrixFile)
        curvatureMatrixFile.close()
        
        # Creation of tail angle graph for bout k
        bStart = superStruct["wellPoissMouv"][i][j][k]["BoutStart"]
        bEnd   = superStruct["wellPoissMouv"][i][j][k]["BoutEnd"]
        tailAngle = superStruct["wellPoissMouv"][i][j][k]["TailAngle_smoothed"]
        if bEnd - bStart + 1 == len(tailAngle):
          plt.figure(1)
          plt.plot([i for i in range(bStart, bEnd + 1)], [t*(180/math.pi) for t in tailAngle])
          if hyperparameters["perBoutOutputYaxis"]:
            plt.axis([bStart, bEnd + 1, hyperparameters["perBoutOutputYaxis"][0], hyperparameters["perBoutOutputYaxis"][1]])
          plt.savefig(os.path.join(outputPath, hyperparameters["videoName"] + "_angle_bout" + str(i) + '_' + str(j) + '_' + str(k) + '.png'))
          plt.close(1)
        
        # Creation of sub-video for bout k
        
        x0 = superStruct["wellPoissMouv"][i][j][k]["HeadX"][0]
        y0 = superStruct["wellPoissMouv"][i][j][k]["HeadY"][0]
        TailX_VideoReferential = superStruct["wellPoissMouv"][i][j][k]["TailX_VideoReferential"]
        TailY_VideoReferential = superStruct["wellPoissMouv"][i][j][k]["TailY_VideoReferential"]
        tailExtX = TailX_VideoReferential[0][len(TailX_VideoReferential[0])-1]
        tailExtY = TailY_VideoReferential[0][len(TailY_VideoReferential[0])-1]
        dist = int((math.sqrt((tailExtX-x0)**2 + (tailExtY-y0)**2))*1.3)
        
        outputName = os.path.join(outputPath, hyperparameters["videoName"] + "_bout" + str(i) + '_' + str(j) + '_' + str(k) + '.avi')
        if dist < cap.get(3) and dist < cap.get(4):
          outputVideoX = int(dist)
          outputVideoY = int(dist)
        else:
          outputVideoX = min(cap.get(3), cap.get(4))
          outputVideoY = min(cap.get(3), cap.get(4))
        
        out = cv2.VideoWriter(outputName,cv2.VideoWriter_fourcc('M','J','P','G'), 10, (outputVideoX, outputVideoY))
        
        BoutStart = superStruct["wellPoissMouv"][i][j][k]["BoutStart"] - perBoutOutputVideoStartStopFrameMargin
        BoutEnd = superStruct["wellPoissMouv"][i][j][k]["BoutEnd"] + perBoutOutputVideoStartStopFrameMargin
        if BoutStart < 0:
          BoutStart = 0
        if BoutEnd >= lastFrame:
          BoutEnd = lastFrame - 1
        mouvLength = len(superStruct["wellPoissMouv"][i][j][k]["HeadX"])
        cap.set(1, BoutStart - firstFrame)
        
        for l in range(BoutStart, BoutEnd):
          ret, frame = cap.read()
          
          if l < superStruct["wellPoissMouv"][i][j][k]["BoutStart"]:
            l2 = 0
          elif l >= superStruct["wellPoissMouv"][i][j][k]["BoutEnd"]:
            l2 = mouvLength - 1
          else:
            l2 = l - superStruct["wellPoissMouv"][i][j][k]["BoutStart"]
          x = superStruct["wellPoissMouv"][i][j][k]["HeadX"][l2]
          y = superStruct["wellPoissMouv"][i][j][k]["HeadY"][l2]
          Heading = superStruct["wellPoissMouv"][i][j][k]["Heading"][l2]
          
          x = int(x + infoWells[i][0])
          y = int(y + infoWells[i][1])
          
          rows = len(frame)
          cols = len(frame[0])
          M = cv2.getRotationMatrix2D((x, y), -(((math.pi/2) - (Heading+math.pi))%(2*math.pi))*(180/math.pi), 1)
          frame = cv2.warpAffine(frame, M, (cols,rows))
          
          if int(x-dist/2)+outputVideoX <= 0:
            print("problem in perBoutOutput")
          else:
            if int(x-dist/2) > 0:
              frame = frame[int(y):int(y)+outputVideoY, int(x-dist/2):int(x-dist/2)+outputVideoX]
            else:
              frame = frame[int(y):int(y)+outputVideoY, 0:outputVideoX]
          
          frame2 = np.zeros((outputVideoX, outputVideoY, 3), np.uint8)
          frame2[0:len(frame), 0:len(frame[0]), :] = frame[0:len(frame), 0:len(frame[0]), :]
          frame = frame2
          
          cv2.putText(frame, str(l + hyperparameters["firstFrame"] - 1), (int(outputVideoX - 100), int(outputVideoY - 30)), cv2.FONT_HERSHEY_SIMPLEX, 1, (0, 255, 0))
          
          out.write(frame)
          
        out.release()



import sys
import os
dir_path = os.path.dirname(os.path.realpath(__file__))
print("The data produced by ZebraZoom can be found in the folder: " + os.path.join(dir_path,'ZZoutput'))

from zebrazoom.code.vars import getGlobalVariables
globalVariables = getGlobalVariables()

if len(sys.argv) == 1:

  from zebrazoom.GUIAllPy import SampleApp
  if __name__ == "__main__":
    app = SampleApp()
    app.mainloop()
    
else:

  if sys.argv[1] == "getTailExtremityFirstFrame":
    
    pathToVideo = sys.argv[2]
    videoName   = sys.argv[3]
    videoExt    = sys.argv[4]
    configFile  = sys.argv[5]
    argv        = sys.argv
    argv.pop(0)
    from zebrazoom.getTailExtremityFirstFrame import getTailExtremityFirstFrame
    if __name__ == '__main__':
      __spec__ = "ModuleSpec(name='builtins', loader=<class '_frozen_importlib.BuiltinImporter'>)"
      getTailExtremityFirstFrame(pathToVideo, videoName, videoExt, configFile, argv)
      
  elif sys.argv[1] == "recreateSuperStruct":
    
    pathToVideo = sys.argv[2]
    videoName   = sys.argv[3]
    videoExt    = sys.argv[4]
    configFile  = sys.argv[5]
    argv        = sys.argv
    argv.pop(0)
    from zebrazoom.recreateSuperStruct import recreateSuperStruct
    if __name__ == '__main__':
      __spec__ = "ModuleSpec(name='builtins', loader=<class '_frozen_importlib.BuiltinImporter'>)"
      recreateSuperStruct(pathToVideo, videoName, videoExt, configFile, argv)
      
  elif sys.argv[1] == "convertSeqToAvi":
    
    from zebrazoom.videoFormatConversion.seq_to_avi import sqb_convert_to_avi
    path      = sys.argv[2]
    videoName = sys.argv[3]
    if len(sys.argv) == 5:
      lastFrame = int(sys.argv[4])
    else:
      lastFrame = -1
    sqb_convert_to_avi(path, videoName, lastFrame)

  else:
  
    pathToVideo = sys.argv[1]
    videoName   = sys.argv[2]
    videoExt    = sys.argv[3]
    configFile  = sys.argv[4]
    argv        = sys.argv
    from zebrazoom.mainZZ import mainZZ
    if __name__ == '__main__':
      __spec__ = "ModuleSpec(name='builtins', loader=<class '_frozen_importlib.BuiltinImporter'>)"
      mainZZ(pathToVideo, videoName, videoExt, configFile, argv)




from zebrazoom.code.dataPostProcessing.generateAllTimeTailAngleGraph import generateAllTimeTailAngleGraph
from zebrazoom.code.dataPostProcessing.perBoutOutput import perBoutOutput
from zebrazoom.code.dataPostProcessing.computeEyesHeadingPlot import computeEyesHeadingPlot

def dataPostProcessing(outputFolderVideo, superStruct, hyperparameters, videoName):

  if hyperparameters["generateAllTimeTailAngleGraph"]:
    generateAllTimeTailAngleGraph(outputFolderVideo, superStruct, hyperparameters["generateAllTimeTailAngleGraphLineWidth"])  
  
  if hyperparameters["perBoutOutput"]:
    perBoutOutput(superStruct, hyperparameters, videoName)
  
  if hyperparameters["computeEyesHeadingPlot"]:
    computeEyesHeadingPlot(superStruct, hyperparameters, videoName)



import numpy as np
import json
import cv2
import math
from zebrazoom.code.getHyperparameters import getHyperparametersSimple
from zebrazoom.code.getImage.getForegroundImage import getForegroundImage
from zebrazoom.mainZZ import mainZZ
import cvui
import pickle
import json
import os
import re
from zebrazoom.code.getBackground import getBackground
from zebrazoom.code.findWells import findWells
from zebrazoom.code.trackingFolder.tracking import tracking
from zebrazoom.code.GUI.adjustParameterInsideAlgoFunctions import prepareConfigFileForParamsAdjustements

def automaticallyFindOptimalParameters(self, controller, realExecThroughGUI, detectBouts, method):
  
  nbOfImagesToManuallyClassify = 3
  headAreaDividedByBodyAreaTarget = 1.1 #0.4
  
  saveIntermediary = True # Should be set to True except when debugging
  
  # # # Getting ground truth from user: head center and tail extremity coordinates for a few frames
  
  if realExecThroughGUI:
    
    videoPath = self.videoToCreateConfigFileFor
    
    for m in re.finditer('/', videoPath):
      last = m.start()
      pathToVideo      = videoPath[:last+1]
      videoNameWithExt = videoPath[last+1:]
      allDotsPositions = [m.start() for m in re.finditer('\.', videoNameWithExt)]
      pointPos  = allDotsPositions[len(allDotsPositions)-1]
      videoName = videoNameWithExt[:pointPos]
      videoExt  = videoNameWithExt[pointPos+1:]
    
    initialConfigFile = self.configFile
    
    initialHyperparameters = getHyperparametersSimple(initialConfigFile)
    initialHyperparameters["videoName"]      = videoName
    wellPositions = findWells(os.path.join(pathToVideo, videoNameWithExt), initialHyperparameters)
    
    cap   = cv2.VideoCapture(videoPath)
    max_l = int(cap.get(7))
    
    # Finding the well with the most movement
    cap.set(1, 0)
    ret, firstFrame = cap.read()
    firstFrame = cv2.cvtColor(firstFrame, cv2.COLOR_BGR2GRAY)
    cap.set(1, max_l-1)
    ret, lastFrame = cap.read()
    lastFrame = cv2.cvtColor(lastFrame, cv2.COLOR_BGR2GRAY)
    pixelsChange = np.zeros((len(wellPositions)))
    for wellNumberId in range(0, len(wellPositions)):
      xtop  = wellPositions[wellNumberId]['topLeftX']
      ytop  = wellPositions[wellNumberId]['topLeftY']
      lenX  = wellPositions[wellNumberId]['lengthX']
      lenY  = wellPositions[wellNumberId]['lengthY']
      firstFrameROI = firstFrame[ytop:ytop+lenY, xtop:xtop+lenX]
      lastFrameROI  = lastFrame[ytop:ytop+lenY, xtop:xtop+lenX]
      pixelsChange[wellNumberId] = np.sum(abs(firstFrameROI-lastFrameROI))
    wellNumber = np.argmax(pixelsChange)
    
    backCalculationStep = int(max_l / nbOfImagesToManuallyClassify)
    data = []
    
    for k in range(0, max_l):
      
      if (k % backCalculationStep == 0):
        
        cap.set(1, k)
        ret, frame = cap.read()
        
        xtop  = wellPositions[wellNumber]['topLeftX']
        ytop  = wellPositions[wellNumber]['topLeftY']
        lenX  = wellPositions[wellNumber]['lengthX']
        lenY  = wellPositions[wellNumber]['lengthY']
        frame = frame[ytop:ytop+lenY, xtop:xtop+lenX]
        
        WINDOW_NAME_1 = "Click on the center of the head of one animal"
        cvui.init(WINDOW_NAME_1)
        cv2.moveWindow(WINDOW_NAME_1, 0,0)
        cvui.imshow(WINDOW_NAME_1, frame)
        while not(cvui.mouse(cvui.CLICK)):
          cursor = cvui.mouse()
          if cv2.waitKey(20) == 27:
            break
          headCoordinates = [cursor.x, cursor.y]
        cv2.destroyAllWindows()
        
        WINDOW_NAME_2 = "Click on the tip of the tail of the same animal"
        cvui.init(WINDOW_NAME_2)
        cv2.moveWindow(WINDOW_NAME_2, 0,0)
        frame2 = frame.copy()
        frame2 = cv2.circle(frame, (headCoordinates[0], headCoordinates[1]), 2, (0, 0, 255), -1)
        cvui.imshow(WINDOW_NAME_2, frame2)
        while not(cvui.mouse(cvui.CLICK)):
          cursor = cvui.mouse()
          if cv2.waitKey(20) == 27:
            break
          tailTipCoordinates = [cursor.x, cursor.y]
        frame2 = cv2.circle(frame, (tailTipCoordinates[0], tailTipCoordinates[1]), 2, (0, 0, 255), -1)  
        cvui.imshow(WINDOW_NAME_2, frame2)
        cv2.waitKey(2000)
        cv2.destroyAllWindows()
        
        if True: # Centered on the animal
          minX = min(headCoordinates[0], tailTipCoordinates[0])
          maxX = max(headCoordinates[0], tailTipCoordinates[0])
          minY = min(headCoordinates[1], tailTipCoordinates[1])
          maxY = max(headCoordinates[1], tailTipCoordinates[1])
          lengthX = maxX - minX
          lengthY = maxY - minY
          
          widdeningFactor = 2
          minX = minX - int(widdeningFactor * lengthX)
          maxX = maxX + int(widdeningFactor * lengthX)
          minY = minY - int(widdeningFactor * lengthY)
          maxY = maxY + int(widdeningFactor * lengthY)
          
          oneWellManuallyChosenTopLeft     = [minX, minY]
          oneWellManuallyChosenBottomRight = [maxX, maxY]
        else: # Focused on the initial well
          oneWellManuallyChosenTopLeft     = [xtop, ytop]
          oneWellManuallyChosenBottomRight = [xtop + lenX, ytop + lenY]          
        
        data.append({"image": frame, "headCoordinates": headCoordinates, "tailTipCoordinates": tailTipCoordinates, "oneWellManuallyChosenTopLeft": oneWellManuallyChosenTopLeft, "oneWellManuallyChosenBottomRight": oneWellManuallyChosenBottomRight, "frameNumber": k, "wellNumber": wellNumber})
  
    if saveIntermediary:
      toSave    = [initialConfigFile, videoPath, data, wellPositions, pathToVideo, videoNameWithExt, videoName, videoExt]
      pickle.dump(toSave, open(videoName + '_paramSet', 'wb'))
      
    cap.release()
    
  else:
    
    toSave = pickle.load(open(self, 'rb'))
    initialConfigFile = toSave[0]
    videoPath         = toSave[1]
    data              = toSave[2]
    wellPositions     = toSave[3]
    pathToVideo       = toSave[4]
    videoNameWithExt  = toSave[5]
    videoName         = toSave[6]
    videoExt          = toSave[7]
    
    # data2 = []
    # data2.append(data[0])
    # data2.append(data[1])
    # data = data2
  
  print("initialConfigFile:", initialConfigFile)
  
  # # # Starting the process of finding the best hyperparameters to track the video
  
  configFile = {"extractAdvanceZebraParameters": 0, "headEmbeded": 0, "nbWells": 1, "noBoutsDetection": 1, "noChecksForBoutSelectionInExtractParams": 1, "trackingPointSizeDisplay": 4, "validationVideoPlotHeading": 1, "nbAnimalsPerWell": 1, "forceBlobMethodForHeadTracking": 1, "multipleHeadTrackingIterativelyRelaxAreaCriteria": 1, "erodeIter":0, "minArea": 0, "maxArea": 100000000000000, "minAreaBody": 0, "maxAreaBody": 100000000000000, "headSize": 20, "minTailSize": 0, "maxTailSize": 100000000000000, "paramGaussianBlur": 25, "extractBackWhiteBackground": 1, "dilateIter": 1, "thresholdForBlobImg": 254, "findContourPrecision": "CHAIN_APPROX_NONE", "midlineIsInBlobTrackingOptimization": 0, "checkAllContourForTailExtremityDetect": 1, "recalculateForegroundImageBasedOnBodyArea": 0, "headingCalculationMethod": "simplyFromPreviousCalculations", "detectMouthInsteadOfHeadTwoSides": 1, "findCenterOfAnimalByIterativelyDilating": 1}
  
  hyperparameters = getHyperparametersSimple(configFile)
  
  background = getBackground(os.path.join(pathToVideo, videoNameWithExt), hyperparameters)
  
  hyperparameters["minPixelDiffForBackExtract"] = 20
  
  # # # For each frame that has some ground truth provided by the user:
  # # # Finds the parameter minPixelDiffForBackExtract (as well as the corresponding number of black pixels in the black frame) that leads to a tail extremity detected as close as possible to the tail extremity value provided by the user
  
  for image in data:
    
    oneWellManuallyChosenTopLeft     = image["oneWellManuallyChosenTopLeft"]
    oneWellManuallyChosenBottomRight = image["oneWellManuallyChosenBottomRight"]
    
    hyperparameters["oneWellManuallyChosenTopLeft"]     = oneWellManuallyChosenTopLeft
    hyperparameters["oneWellManuallyChosenBottomRight"] = oneWellManuallyChosenBottomRight
    hyperparameters["videoWidth"]    = oneWellManuallyChosenBottomRight[0] - oneWellManuallyChosenTopLeft[0]
    hyperparameters["videoHeight"]   = oneWellManuallyChosenBottomRight[1] - oneWellManuallyChosenTopLeft[1]
    
    hyperparameters["firstFrame"]    = image["frameNumber"]
    hyperparameters["lastFrame"]     = image["frameNumber"]
    hyperparameters["debugTracking"] = 0
    
    hyperparameters["fixedHeadPositionX"] = int(image["headCoordinates"][0])
    hyperparameters["fixedHeadPositionY"] = int(image["headCoordinates"][1])
    hyperparameters["midlineIsInBlobTrackingOptimization"] = 0
    
    bestMinPixelDiffForBackExtract = 10
    lowestTailTipDistError = 1000000000
    for minPixelDiffForBackExtract in range(3, 25):
      hyperparameters["minPixelDiffForBackExtract"] = minPixelDiffForBackExtract
      trackingData = tracking(videoPath, background, image["wellNumber"], wellPositions, hyperparameters, videoName)
      tailTipGroundTruth = image["tailTipCoordinates"] #[image["tailTipCoordinates"][0] - image["oneWellManuallyChosenTopLeft"][0], image["tailTipCoordinates"][1] - image["oneWellManuallyChosenTopLeft"][1]]
      tailTipPredicted = trackingData[0][0][0][len(trackingData[0][0][0])-1]
      if (trackingData[0][0][0][0][0] == tailTipPredicted[0] and trackingData[0][0][0][0][1] == tailTipPredicted[1]) or (trackingData[0][0][0][1][0] == tailTipPredicted[0] and trackingData[0][0][0][1][1] == tailTipPredicted[1]):
        tailTipDistError = 1000000000
      else:
        tailTipDistError = math.sqrt((tailTipGroundTruth[0] - tailTipPredicted[0])** 2 + (tailTipGroundTruth[1] - tailTipPredicted[1])**2)
      if tailTipDistError < lowestTailTipDistError:
        lowestTailTipDistError = tailTipDistError
        bestMinPixelDiffForBackExtract = minPixelDiffForBackExtract
        previousDataPoint = []
        tailLength = 0
        for dataPoint in trackingData[0][0][0]:
          if len(previousDataPoint):
            tailLength = tailLength + math.sqrt((dataPoint[0] - previousDataPoint[0])**2 + (dataPoint[1] - previousDataPoint[1])**2)
          previousDataPoint = dataPoint
        image["tailLength"] = tailLength
        image["tailLengthManual"] = math.sqrt((image["headCoordinates"][0] - tailTipGroundTruth[0])**2 + (image["headCoordinates"][1] - tailTipGroundTruth[1])**2)
      print("minPixelDiffForBackExtract:", minPixelDiffForBackExtract, "; tailTipDistError:", tailTipDistError)
    
    image["bestMinPixelDiffForBackExtract"] = bestMinPixelDiffForBackExtract
    image["lowestTailTipDistError"]         = lowestTailTipDistError
    
    if lowestTailTipDistError != 1000000000:
      hyperparameters["minPixelDiffForBackExtract"] = image["bestMinPixelDiffForBackExtract"]
      [foregroundImage, o1, o2] = getForegroundImage(videoPath, background, image["frameNumber"], image["wellNumber"], wellPositions, hyperparameters)
      ret, thresh = cv2.threshold(foregroundImage, hyperparameters["thresholdForBlobImg"], 255, cv2.THRESH_BINARY)
      thresh[0,:] = 255
      thresh[len(thresh)-1,:] = 255
      thresh[:,0] = 255
      thresh[:,len(thresh[0])-1] = 255
      contourClickedByUser = 0
      # cv2.imshow('Frame', thresh)
      # cv2.waitKey(0)
      contours, hierarchy = cv2.findContours(thresh, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)
      for contour in contours:
        dist = cv2.pointPolygonTest(contour, (image["headCoordinates"][0], image["headCoordinates"][1]), True)
        if dist >= 0:
          contourClickedByUser = contour
      if not(type(contourClickedByUser) == int): # We found the contour that the user selected
        bodyContourArea = cv2.contourArea(contourClickedByUser)
        image["bodyContourArea"] = bodyContourArea
      else:
        image["bodyContourArea"] = -1
      
    del hyperparameters["fixedHeadPositionX"]
    del hyperparameters["fixedHeadPositionY"]
    del hyperparameters["midlineIsInBlobTrackingOptimization"]
    
  # # # Choosing the hyperparameters related to the background substraction and to the tail length as well as the headSize hyperparameter (for the heading calculation)
  
  configFile["nbAnimalsPerWell"]      = initialConfigFile["nbAnimalsPerWell"]
  hyperparameters["nbAnimalsPerWell"] = initialConfigFile["nbAnimalsPerWell"]
  
  tailLengthManualOptions = []
  for image in data:
    tailLengthManualOptions.append(image["tailLengthManual"])
  maxTailLengthManual = max(tailLengthManualOptions)
  print("tailLengthManualOptions:", tailLengthManualOptions)
  print("maxTailLengthManual:", maxTailLengthManual)
  
  bestMinPixelDiffForBackExtract = -1
  bestMinPixelDiffForBackExtractOptions = []
  bodyContourAreaOptions = []
  tailLengthOptions = []
  for image in data:
    if image["lowestTailTipDistError"] != 1000000000 and (image["tailLength"] < 10 * maxTailLengthManual):
      if image["bodyContourArea"] != -1:
        bestMinPixelDiffForBackExtractOptions.append(image["bestMinPixelDiffForBackExtract"])
        tailLengthOptions.append(image["tailLength"])
        bodyContourAreaOptions.append(image["bodyContourArea"])
  
  print("bestMinPixelDiffForBackExtractOptions:", bestMinPixelDiffForBackExtractOptions)
  print("tailLengthOptions:", tailLengthOptions)
  print("bodyContourAreaOptions:", bodyContourAreaOptions)
  
  if len(bestMinPixelDiffForBackExtractOptions):
    if False:
      bestMinPixelDiffForBackExtract = int(np.max(bestMinPixelDiffForBackExtractOptions))
      tailLength = int(np.min(tailLengthOptions))
      bodyContourArea = int(np.min(bodyContourAreaOptions))
    else:
      ind = np.argmin(bodyContourAreaOptions)
      bestMinPixelDiffForBackExtract = bestMinPixelDiffForBackExtractOptions[ind]
      tailLength      = tailLengthOptions[ind]
      bodyContourArea = bodyContourAreaOptions[ind]
  else:
    print("NEED TO START OVER: bodyContourArea")
    bodyContourArea = 0
  if bestMinPixelDiffForBackExtract == -1:
    print("NEED TO START OVER")
  
  hyperparameters["minPixelDiffForBackExtract"] = bestMinPixelDiffForBackExtract
  hyperparameters["minTailSize"] = tailLength / 10
  hyperparameters["maxTailSize"] = tailLength * 2
  hyperparameters["headSize"]    = int(tailLength / 2)
  hyperparameters["adjustMinPixelDiffForBackExtract_nbBlackPixelsMax"] = bodyContourArea * configFile["nbAnimalsPerWell"]
  hyperparameters["maxAreaBody"] = 5 * bodyContourArea
  hyperparameters["minAreaBody"] = int(bodyContourArea / 5)
  hyperparameters["trackingPointSizeDisplay"] = int(np.ceil(math.sqrt(bodyContourArea) / 15))
  configFile["minPixelDiffForBackExtract"] = bestMinPixelDiffForBackExtract
  configFile["minTailSize"] = tailLength / 10
  configFile["maxTailSize"] = tailLength * 2
  configFile["headSize"]    = int(tailLength / 2)
  configFile["adjustMinPixelDiffForBackExtract_nbBlackPixelsMax"] = bodyContourArea * configFile["nbAnimalsPerWell"]
  configFile["maxAreaBody"] = 5 * bodyContourArea
  configFile["minAreaBody"] = int(bodyContourArea / 5)
  configFile["trackingPointSizeDisplay"] = int(np.ceil(math.sqrt(bodyContourArea) / 15))
  
  # # # For each frame that has some ground truth provided by the user:
  # # # Finds the area of the body as well as the optimal dilateIter hyperparameters to go from body contour to head contour
  
  for image in data:
    
    if image["lowestTailTipDistError"] != 1000000000 and image["bodyContourArea"] and (image["tailLength"] < 10 * maxTailLengthManual):
      
      bodyContourArea = image["bodyContourArea"]
      
      [foregroundImage, o1, o2] = getForegroundImage(videoPath, background, image["frameNumber"], image["wellNumber"], wellPositions, hyperparameters)
      
      ret, thresh = cv2.threshold(foregroundImage, hyperparameters["thresholdForBlobImg"], 255, cv2.THRESH_BINARY)
      thresh[0,:] = 255
      thresh[len(thresh)-1,:] = 255
      thresh[:,0] = 255
      thresh[:,len(thresh[0])-1] = 255
      contourClickedByUser = 0
      contours, hierarchy = cv2.findContours(thresh, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)
      for contour in contours:
        dist = cv2.pointPolygonTest(contour, (image["headCoordinates"][0], image["headCoordinates"][1]), True)
        if dist >= 0:
          contourClickedByUser = contour
      
      if not(type(contourClickedByUser) == int): # We found the contour that the user selected
        
        # bodyContourArea = cv2.contourArea(contourClickedByUser) # New
        
        image["contourClickedByUser"] = contourClickedByUser
        thresh2 = thresh.copy()
        thresh2[0,:] = 255
        thresh2[len(thresh2)-1,:] = 255
        thresh2[:,0] = 255
        thresh2[:,len(thresh2[0])-1] = 255
        erodeSize = hyperparameters["erodeSize"]
        kernel  = np.ones((erodeSize, erodeSize), np.uint8)
        
        curHeadAreaDividedByBodyArea = 1.1
        dilateIter = 0
        while curHeadAreaDividedByBodyArea > headAreaDividedByBodyAreaTarget:
          dilateIter = dilateIter + 1
          thresh2_b = cv2.dilate(thresh2, kernel, iterations=dilateIter)
          contours, hierarchy = cv2.findContours(thresh2_b, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)
          correspondingHeadContour = 0
          maxDist    = 100000000
          for contour in contours:
            M = cv2.moments(contour)
            if M['m00']:
              x = int(M['m10']/M['m00'])
              y = int(M['m01']/M['m00'])
            else:
              x = 0
              y = 0
            dist = cv2.pointPolygonTest(contourClickedByUser, (x, y), True)
            if dist >= 0:
              dist = math.sqrt((image["headCoordinates"][0] - x)**2 + (image["headCoordinates"][1] - y)**2)
              if dist < maxDist:
                maxDist = dist
                correspondingHeadContour = contour
          if not(type(correspondingHeadContour) == int):
            headContourArea = cv2.contourArea(correspondingHeadContour)
            curHeadAreaDividedByBodyArea = headContourArea / bodyContourArea
            print("curHeadAreaDividedByBodyArea, headContourArea, bodyContourArea, dilateIter:", curHeadAreaDividedByBodyArea, headContourArea, bodyContourArea, dilateIter)
          else:
            curHeadAreaDividedByBodyArea = 0
        if curHeadAreaDividedByBodyArea == 0:
          dilateIter = dilateIter - 1
        image["dilateIter"] = dilateIter
  
  # # # Choosing the hyperparameters related to the max and min of the body contour as well as the dilateIter hyperparameters to go from body contour to head contour
  
  dilateIterOptions = []
  for image in data:
    if image["lowestTailTipDistError"] != 1000000000 and (image["tailLength"] < 10 * maxTailLengthManual) and ("dilateIter" in image):
      dilateIterOptions.append(image["dilateIter"])
      
  print("dilateIterOptions:", dilateIterOptions)
  
  if len(dilateIterOptions):
    hyperparameters["dilateIter"] = int(np.ceil(np.median(dilateIterOptions)))
    configFile["dilateIter"] = int(np.ceil(np.median(dilateIterOptions)))
  else:
    print("NEED TO START OVER FOR dilateIter")
  
  # # # For each frame that has some ground truth provided by the user:
  # # # Finds the area of the head
  
  for image in data:
  
    if image["lowestTailTipDistError"] != 1000000000 and (image["tailLength"] < 10 * maxTailLengthManual) and ("contourClickedByUser" in image):
    
      [foregroundImage, o1, o2] = getForegroundImage(videoPath, background, image["frameNumber"], image["wellNumber"], wellPositions, hyperparameters)
      ret, thresh = cv2.threshold(foregroundImage, hyperparameters["thresholdForBlobImg"], 255, cv2.THRESH_BINARY)
      thresh[0,:] = 255
      thresh[len(thresh)-1,:] = 255
      thresh[:,0] = 255
      thresh[:,len(thresh[0])-1] = 255
      contourClickedByUser = image["contourClickedByUser"]
      
      thresh2 = cv2.dilate(thresh, kernel, iterations=hyperparameters["dilateIter"])
      contours, hierarchy = cv2.findContours(thresh2, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)
      correspondingHeadContour = 0
      maxDist    = 100000000
      for contour in contours:
        M = cv2.moments(contour)
        if M['m00']:
          x = int(M['m10']/M['m00'])
          y = int(M['m01']/M['m00'])
        else:
          x = 0
          y = 0
        dist = cv2.pointPolygonTest(contourClickedByUser, (x, y), True)
        if dist >= 0:
          dist = math.sqrt((image["headCoordinates"][0] - x)**2 + (image["headCoordinates"][1] - y)**2)
          if dist < maxDist:
            maxDist = dist
            correspondingHeadContour = contour
      
      if not(type(correspondingHeadContour) == int):
        headContourArea = cv2.contourArea(correspondingHeadContour)
        image["headContourArea"] = headContourArea
  
  # # # Choosing the hyperparameters related to the max and min of the head contour
  
  headContourAreaOptions = []
  for image in data:
    if image["lowestTailTipDistError"] != 1000000000 and (image["tailLength"] < 10 * maxTailLengthManual):
      if "headContourArea" in image:
        headContourAreaOptions.append(image["headContourArea"])
  print("headContourAreaOptions:", headContourAreaOptions)
  if len(headContourAreaOptions):
    headContourArea = np.mean(headContourAreaOptions)
  else:
    print("Start Over!!!")
  hyperparameters["minArea"] = 0.7 * headContourArea
  hyperparameters["maxArea"] = 1.3 * headContourArea
  configFile["minArea"] = 0.7 * headContourArea
  configFile["maxArea"] = 1.3 * headContourArea
  
  # # # Overwritting some of the hyperparameters from hyperparameters in the initial configFile
  
  listOfParametersToOverwrite = ["nbWells", "nbRowsOfWells", "nbWellsPerRows", "minWellDistanceForWellDetection", "wellOutputVideoDiameter", "wellsAreRectangles", "rectangleWellAreaImageThreshold", "rectangleWellErodeDilateKernelSize", "findRectangleWellArea", "rectangularWellsInvertBlackWhite", "noWellDetection", "oneWellManuallyChosenTopLeft", "oneWellManuallyChosenBottomRight"]
  for parameter in listOfParametersToOverwrite:
    if parameter in initialConfigFile:
      configFile[parameter] = initialConfigFile[parameter]
      
  if method:
    configFile["recalculateForegroundImageBasedOnBodyArea"] = 1
  
  print("Intermediary config file:", configFile)
  
  if configFile["nbAnimalsPerWell"] == 1:
  
    if detectBouts:
    
      # Extracting Background and finding Wells
      
      configFile["exitAfterBackgroundExtraction"] = 1
      configFile["debugExtractBack"]              = 1
      configFile["debugFindWells"]                = 1
      
      WINDOW_NAME = "Please Wait"
      cvui.init(WINDOW_NAME)
      cv2.moveWindow(WINDOW_NAME, 0,0)
      img = np.zeros((400, 900, 3), np.uint8)
      lineType               = 2
      img = cv2.putText(img,'In the next window, you will need to adjust parameters in order', (5, 100), cv2.FONT_HERSHEY_SIMPLEX, 0.7, (255,255,255), lineType)
      cv2.putText(img,'for the red dot to appear when and only when the animal is moving.', (5, 150), cv2.FONT_HERSHEY_SIMPLEX, 0.7, (255,255,255), lineType)
      cv2.imshow(WINDOW_NAME, img)
      r = cv2.waitKey(10)
      
      try:
        mainZZ(pathToVideo, videoName, videoExt, configFile, [])
      except ValueError:
        configFile["exitAfterBackgroundExtraction"] = 0
      
      cv2.destroyAllWindows()
      
      del configFile["exitAfterBackgroundExtraction"]
      del configFile["debugExtractBack"]
      del configFile["debugFindWells"]
    
      # Finding the frame with the most movement
      
      cap   = cv2.VideoCapture(videoPath)
      max_l = int(cap.get(7))
      
      wellNumber = data[0]["wellNumber"]
      xtop = wellPositions[wellNumber]['topLeftX']
      ytop = wellPositions[wellNumber]['topLeftY']
      lenX = wellPositions[wellNumber]['lengthX']
      lenY = wellPositions[wellNumber]['lengthY']
      
      print("xtop, ytop, lenX, lenY:", xtop, ytop, lenX, lenY)
      
      firstFrameNum  = 0
      lastFrameNum   = max_l - 1
      
      while abs(lastFrameNum - firstFrameNum) > 40:
        
        middleFrameNum = int((firstFrameNum + lastFrameNum) / 2)
        
        cap.set(1, firstFrameNum)
        ret, firstFrame = cap.read()
        firstFrame = cv2.cvtColor(firstFrame, cv2.COLOR_BGR2GRAY)
        
        cap.set(1, middleFrameNum)
        ret, middleFrame = cap.read()
        middleFrame = cv2.cvtColor(middleFrame, cv2.COLOR_BGR2GRAY)
        
        cap.set(1, lastFrameNum)
        ret, lastFrame = cap.read()
        lastFrame = cv2.cvtColor(lastFrame, cv2.COLOR_BGR2GRAY)
        
        firstFrameROI  = firstFrame[ytop:ytop+lenY, xtop:xtop+lenX]
        middleFrameROI = middleFrame[ytop:ytop+lenY, xtop:xtop+lenX]
        lastFrameROI   = lastFrame[ytop:ytop+lenY, xtop:xtop+lenX]
        pixelsChangeFirstToMiddle = np.sum(abs(firstFrameROI-middleFrameROI))
        pixelsChangeMiddleToLast  = np.sum(abs(middleFrameROI-lastFrameROI))
        
        if pixelsChangeFirstToMiddle > pixelsChangeMiddleToLast:
          lastFrameNum  = middleFrameNum
        else:
          firstFrameNum = middleFrameNum
        
        print("intermediary: firstFrameNum:", firstFrameNum, "; lastFrameNum:", lastFrameNum)
        print("pixelsChangeFirstToMiddle:", pixelsChangeFirstToMiddle, "; pixelsChangeMiddleToLast:", pixelsChangeMiddleToLast)
        
      cap.release()
      
      print("wellNumber:", wellNumber, "; lastFrameNum:", lastFrameNum, "; firstFrameNum:", firstFrameNum)
      
      # Launching the interactive adjustement of hyperparameters related to the detection of bouts
      
      wellNumber = str(data[0]["wellNumber"])
      firstFrameParamAdjust = 0
      adjustOnWholeVideo    = 0
      
      [configFile, initialFirstFrameValue, initialLastFrameValue] = prepareConfigFileForParamsAdjustements(configFile, wellNumber, firstFrameParamAdjust, videoPath, adjustOnWholeVideo)
      configFile["firstFrame"] = lastFrameNum - 500 if lastFrameNum - 500 > 0 else 0
      configFile["lastFrame"]  = (configFile["firstFrame"] + 500) if (configFile["firstFrame"] + 500) < (max_l - 1) else (max_l - 1)
      
      configFile["noBoutsDetection"] = 0
      configFile["trackTail"]                   = 0
      configFile["adjustDetectMovWithRawVideo"] = 1
      configFile["reloadWellPositions"]         = 1
      
      if "thresForDetectMovementWithRawVideo" in configFile:
        if configFile["thresForDetectMovementWithRawVideo"] == 0:
          configFile["thresForDetectMovementWithRawVideo"] = 1
      else:
        configFile["thresForDetectMovementWithRawVideo"] = 1
      
      try:
        WINDOW_NAME = "Please Wait"
        cvui.init(WINDOW_NAME)
        cv2.moveWindow(WINDOW_NAME, 0,0)
        img = np.zeros((400, 900, 3), np.uint8)
        lineType               = 2
        img = cv2.putText(img,'In the next window, you will need to adjust parameters in order', (5, 100), cv2.FONT_HERSHEY_SIMPLEX, 0.7, (255,255,255), lineType)
        cv2.putText(img,'for the red dot to appear when and only when the animal is moving.', (5, 150), cv2.FONT_HERSHEY_SIMPLEX, 0.7, (255,255,255), lineType)
        cv2.imshow(WINDOW_NAME, img)
        r = cv2.waitKey(10)
        mainZZ(pathToVideo, videoName, videoExt, configFile, [])
        cv2.destroyAllWindows()
      except ValueError:
        newhyperparameters = pickle.load(open('newhyperparameters', 'rb'))
        for index in newhyperparameters:
          configFile[index] = newhyperparameters[index]
      
      del configFile["onlyTrackThisOneWell"]
      del configFile["trackTail"]
      del configFile["adjustDetectMovWithRawVideo"]
      del configFile["reloadWellPositions"]
      
      if initialLastFrameValue == -1:
        if 'firstFrame' in configFile:
          del configFile['firstFrame']
        if 'lastFrame' in configFile:
          del configFile['lastFrame']
      else:
        configFile["firstFrame"]                  = initialFirstFrameValue
        configFile["lastFrame"]                   = initialLastFrameValue
      
      del configFile["reloadBackground"]
      
      configFile["fillGapFrameNb"] = 2 # We would need to try to improve this in the future
  
  # # # Moving on to the next step
  
  if realExecThroughGUI:
    self.configFile = configFile
    controller.show_frame("FinishConfig")
  
  print("final Config File", configFile)
  
  reference = videoName + '_config.json'
  with open(reference, 'w') as outfile:
    json.dump(configFile, outfile)
  
  


import h5py
import numpy as np
import cv2
from zebrazoom.code.popUpAlgoFollow import prepend
import math
import os

from zebrazoom.code.trackingFolder.getImages import getImages
from zebrazoom.code.trackingFolder.headTrackingHeadingCalculationFolder.headTrackingHeadingCalculation import headTrackingHeadingCalculation
from zebrazoom.code.trackingFolder.debugTracking import debugTracking
from zebrazoom.code.trackingFolder.tailTracking import tailTracking
from zebrazoom.code.trackingFolder.blackFramesDetection import getThresForBlackFrame, savingBlackFrames
from zebrazoom.code.trackingFolder.tailTrackingFunctionsFolder.getTailTipManual import getHeadPositionByFileSaved, findTailTipByUserInput, getTailTipByFileSaved, findHeadPositionByUserInput
from zebrazoom.code.trackingFolder.eyeTracking.eyeTracking import eyeTracking
from zebrazoom.code.trackingFolder.postProcessMultipleTrajectories import postProcessMultipleTrajectories

from zebrazoom.code.getImage.headEmbededFrame import headEmbededFrame
from zebrazoom.code.getImage.headEmbededFrameBackExtract import headEmbededFrameBackExtract

from zebrazoom.code.adjustHyperparameters import initializeAdjustHyperparametersWindows, adjustHyperparameters, getHeadEmbededTrackingParamsForHyperParamAdjusts, getFreelySwimTrackingParamsForHyperParamAdjusts
from zebrazoom.code.trackingFolder.tailTrackingFunctionsFolder.headEmbededTailTracking import adjustHeadEmbededHyperparameters

from zebrazoom.code.trackingFolder.tailTrackingFunctionsFolder.headEmbededTailTracking import headEmbededTailTrackFindMaxDepth
from zebrazoom.code.trackingFolder.tailTrackingFunctionsFolder.headEmbededTailTrackingTeresaNicolson import headEmbededTailTrackFindMaxDepthTeresaNicolson, headEmbededTailTrackingTeresaNicolson
from zebrazoom.code.trackingFolder.tailTrackingFunctionsFolder.centerOfMassTailTracking import centerOfMassTailTrackFindMaxDepth

def tracking(videoPath, background, wellNumber, wellPositions, hyperparameters, videoName):

  firstFrame = hyperparameters["firstFrame"]
  if hyperparameters["firstFrameForTracking"] != -1:
    firstFrame = hyperparameters["firstFrameForTracking"]
  lastFrame = hyperparameters["lastFrame"]
  
  if (hyperparameters["freqAlgoPosFollow"] != 0):
    print("lastFrame:",lastFrame)

  nbTailPoints = hyperparameters["nbTailPoints"]
  thetaDiffAccept = 1.2 # 0.5 for the head embedded maybe
  maxDepth = 0
  headPositionFirstFrame = []
  tailTipFirstFrame = []
  
  cap = cv2.VideoCapture(videoPath)
  if (cap.isOpened()== False): 
    print("Error opening video stream or file")
  frame_width  = int(cap.get(3))
  frame_height = int(cap.get(4))
  
  heading = -1
  if hyperparameters["headEmbeded"]:
    heading = 0.7
    
  trackingHeadTailAllAnimals = np.zeros((hyperparameters["nbAnimalsPerWell"], lastFrame-firstFrame+1, nbTailPoints, 2))
  trackingHeadingAllAnimals = np.zeros((hyperparameters["nbAnimalsPerWell"], lastFrame-firstFrame+1))
  if hyperparameters["eyeTracking"]:
    trackingEyesAllAnimals = np.zeros((hyperparameters["nbAnimalsPerWell"], lastFrame-firstFrame+1, 8))
  else:
    trackingEyesAllAnimals = 0
  
  threshForBlackFrames = getThresForBlackFrame(hyperparameters, videoPath) # For headEmbededTeresaNicolson 
  cap.set(1, firstFrame)
  
  # Using the first frame of the video to calculate parameters that will be used afterwards for the tracking
  if (hyperparameters["headEmbeded"] == 1):
    # Getting images
    
    if hyperparameters["headEmbededRemoveBack"] == 0 and hyperparameters["headEmbededAutoSet_BackgroundExtractionOption"] == 0:
      [frame, thresh1] = headEmbededFrame(videoPath, firstFrame, hyperparameters)
    else:
      hyperparameters["headEmbededRemoveBack"] = 1
      hyperparameters["minPixelDiffForBackExtract"] = hyperparameters["headEmbededAutoSet_BackgroundExtractionOption"]
      [frame, thresh1] = headEmbededFrameBackExtract(videoPath, background, hyperparameters, firstFrame)
      
    # if hyperparameters["invertBlackWhiteOnImages"]:
      # frame   = 255 - frame
    
    gray = frame.copy()
    
    oppHeading = (heading + math.pi) % (2 * math.pi)
    
    # Getting headPositionFirstFrame and tailTipFirstFrame positions
    if os.path.exists(videoPath+'HP.csv'):
      headPositionFirstFrame = getHeadPositionByFileSaved(videoPath)
    else:
      if hyperparameters["findHeadPositionByUserInput"]:
        headPositionFirstFrame = findHeadPositionByUserInput(frame, firstFrame, videoPath)
      else:
        [frame, gray, thresh1, blur, thresh2, frame2, initialCurFrame, back] = getImages(hyperparameters, cap, videoPath, firstFrame, background, wellNumber, wellPositions)
        cap.set(1, firstFrame)
        [trackingHeadingAllAnimals, trackingHeadTailAllAnimals, lastFirstTheta] = headTrackingHeadingCalculation(hyperparameters, firstFrame, firstFrame, blur, thresh1, thresh2, gray, hyperparameters["erodeSize"], frame_width, frame_height, trackingHeadingAllAnimals, trackingHeadTailAllAnimals, headPositionFirstFrame, wellPositions[wellNumber]["lengthX"])
    if os.path.exists(videoPath+'.csv'):
      tailTipFirstFrame  = getTailTipByFileSaved(hyperparameters,videoPath)
    else:
      tailTipFirstFrame  = findTailTipByUserInput(frame, firstFrame, videoPath, hyperparameters)
    if hyperparameters["automaticallySetSomeOfTheHeadEmbededHyperparameters"] == 1:
      hyperparameters = adjustHeadEmbededHyperparameters(hyperparameters, frame, headPositionFirstFrame, tailTipFirstFrame)
    # Getting max depth
    if hyperparameters["headEmbededTeresaNicolson"] == 1:
      if len(headPositionFirstFrame) == 0:
        headPositionFirstFrame = [trackingHeadTailAllAnimals[0][0][0][0], trackingHeadTailAllAnimals[0][0][0][1]]
      maxDepth = headEmbededTailTrackFindMaxDepthTeresaNicolson(headPositionFirstFrame,nbTailPoints,firstFrame,headPositionFirstFrame[0],headPositionFirstFrame[1],thresh1,frame,hyperparameters,oppHeading,tailTipFirstFrame)
    else:
      if hyperparameters["centerOfMassTailTracking"] == 0:
        maxDepth = headEmbededTailTrackFindMaxDepth(headPositionFirstFrame,nbTailPoints,firstFrame,headPositionFirstFrame[0],headPositionFirstFrame[1],thresh1,frame,hyperparameters,oppHeading,tailTipFirstFrame)
      else:
        maxDepth = centerOfMassTailTrackFindMaxDepth(headPositionFirstFrame,nbTailPoints,firstFrame,headPositionFirstFrame[0],headPositionFirstFrame[1],thresh1,frame,hyperparameters,oppHeading,tailTipFirstFrame)
  
  if hyperparameters["adjustHeadEmbededTracking"] == 1 or hyperparameters["adjustFreelySwimTracking"] == 1:
    initializeAdjustHyperparametersWindows("Tracking")
  organizationTabCur = []
  
  # Performing the tracking on each frame
  i = firstFrame
  while (i < lastFrame+1):
    
    if (hyperparameters["freqAlgoPosFollow"] != 0) and (i % hyperparameters["freqAlgoPosFollow"] == 0):
      print("Tracking: wellNumber:",wellNumber," ; frame:",i)
      if hyperparameters["popUpAlgoFollow"]:
        prepend("Tracking: wellNumber:" + str(wellNumber) + " ; frame:" + str(i))
    if hyperparameters["debugTracking"]:
      print("frame:",i)
    # Get images for frame i
    [frame, gray, thresh1, blur, thresh2, frame2, initialCurFrame, back] = getImages(hyperparameters, cap, videoPath, i, background, wellNumber, wellPositions)
    # Head tracking and heading calculation
    [trackingHeadingAllAnimals, trackingHeadTailAllAnimals, lastFirstTheta] = headTrackingHeadingCalculation(hyperparameters, firstFrame, i, blur, thresh1, thresh2, gray, hyperparameters["erodeSize"], frame_width, frame_height, trackingHeadingAllAnimals, trackingHeadTailAllAnimals, headPositionFirstFrame, wellPositions[wellNumber]["lengthX"])
    # Tail tracking for frame i
    if hyperparameters["trackTail"] == 1 :
      for animalId in range(0, hyperparameters["nbAnimalsPerWell"]):
        [trackingHeadTailAllAnimals, trackingHeadingAllAnimals] = tailTracking(animalId, i, firstFrame, videoPath, frame, hyperparameters, thresh1, nbTailPoints, threshForBlackFrames, thetaDiffAccept, trackingHeadTailAllAnimals, trackingHeadingAllAnimals, lastFirstTheta, maxDepth, tailTipFirstFrame, initialCurFrame, back)
    # Eye tracking for frame i
    if hyperparameters["eyeTracking"]:
      trackingEyesAllAnimals = eyeTracking(animalId, i, firstFrame, frame, hyperparameters, thresh1, trackingHeadingAllAnimals, trackingHeadTailAllAnimals, trackingEyesAllAnimals)
    # Debug functions
    debugTracking(nbTailPoints, i, firstFrame, trackingHeadTailAllAnimals, trackingHeadingAllAnimals, frame2, hyperparameters)
    
    if hyperparameters["adjustHeadEmbededTracking"] == 1:
      [hyperparametersListNames, frameToShow, WINDOW_NAME, organizationTab] = getHeadEmbededTrackingParamsForHyperParamAdjusts(nbTailPoints, i, firstFrame, trackingHeadTailAllAnimals, trackingHeadingAllAnimals, frame, frame2, hyperparameters)
      if len(organizationTabCur) == 0:
        organizationTabCur = organizationTab
      [i, hyperparameters, organizationTabCur] = adjustHyperparameters(i, hyperparameters, hyperparametersListNames, frameToShow, WINDOW_NAME, organizationTabCur)
      hyperparameters = adjustHeadEmbededHyperparameters(hyperparameters, frame, headPositionFirstFrame, tailTipFirstFrame)
    elif hyperparameters["adjustFreelySwimTracking"] == 1:
      [hyperparametersListNames, frameToShow, WINDOW_NAME, organizationTab] = getFreelySwimTrackingParamsForHyperParamAdjusts(nbTailPoints, i, firstFrame, trackingHeadTailAllAnimals, trackingHeadingAllAnimals, frame, frame2, hyperparameters)
      if len(organizationTabCur) == 0:
        organizationTabCur = organizationTab
      [i, hyperparameters, organizationTabCur] = adjustHyperparameters(i, hyperparameters, hyperparametersListNames, frameToShow, WINDOW_NAME, organizationTabCur)
    else:
      i = i + 1
      
  if hyperparameters["postProcessMultipleTrajectories"]:
    [trackingHeadingAllAnimals, trackingHeadTailAllAnimals, trackingEyesAllAnimals] = postProcessMultipleTrajectories(trackingHeadingAllAnimals, trackingHeadTailAllAnimals, trackingEyesAllAnimals, hyperparameters)
  
  savingBlackFrames(hyperparameters, videoName, trackingHeadTailAllAnimals)
  
  print("Tracking done for well", wellNumber)
  if hyperparameters["popUpAlgoFollow"]:
    prepend("Tracking done for well "+ str(wellNumber))
  
  return [trackingHeadTailAllAnimals, trackingHeadingAllAnimals, trackingEyesAllAnimals, headPositionFirstFrame, tailTipFirstFrame]



from pathlib import Path
import json
import cv2
import os
from zebrazoom.code.vars import getGlobalVariables
globalVariables = getGlobalVariables()

cur_dir_path = os.path.dirname(os.path.realpath(__file__))

cur_dir_path2 = Path(cur_dir_path)
cur_dir_path2 = cur_dir_path2.parent

with open(os.path.join(cur_dir_path, 'defaultConfigFile.json')) as f:
  configDefault = json.load(f)

# Returns parameters in config file
def getConfig(config, variableName, videoPath):
  
  if variableName in config:
    
    return config[variableName]
    
  else:
    
    if variableName in configDefault:
      
      if variableName == "outputFolder":
        return os.path.join(cur_dir_path2, "ZZoutput")
      else:
        return configDefault[variableName]
        
    else:
      
      cap = 0
      if len(videoPath):
        cap = cv2.VideoCapture(videoPath)
      if (len(videoPath) == 0) or (cap.isOpened() == False):
        print("Error opening video stream or file in getConfig")
        if variableName == "firstFrame":
          return 1
        else:
          return 0
      else:
        if variableName == "firstFrame":
          return 1
        elif variableName == "lastFrame":
          lastFrame = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
          return lastFrame
        elif variableName == "videoWidth":
          frame_width  = int(cap.get(3))
          return frame_width
        elif variableName == "videoHeight":
          frame_height = int(cap.get(4))
          return frame_height
        else:
          print("You need to put the parameter",variableName,"in your configuration file")
          return 0

def getHyperparameters(configFile, videoName, videoPath, argv):

  hyperparameters = {}
  
  if type(configFile) == str:
    with open(configFile) as f:
      config = json.load(f)
  else:
    config = configFile
    
  if "parentConfigFiles" in config:
    for parentConfigFile in config["parentConfigFiles"]:
      with open(parentConfigFile) as f2:
        parentConfig = json.load(f2)
        parentConfig.update(config)
        config = parentConfig
  
  hyperparameters["minWellDistanceForWellDetection"] = getConfig(config, "minWellDistanceForWellDetection", videoPath)
  hyperparameters["wellOutputVideoDiameter"] = getConfig(config, "wellOutputVideoDiameter", videoPath)
  hyperparameters["firstFrame"] = getConfig(config, "firstFrame", videoPath)
  hyperparameters["lastFrame"] = getConfig(config, "lastFrame", videoPath)
  hyperparameters["nbWells"] = getConfig(config, "nbWells", videoPath)
  hyperparameters["backCalculationStep"] = getConfig(config, "backCalculationStep", videoPath)
  hyperparameters["onlyTrackThisOneWell"] = getConfig(config, "onlyTrackThisOneWell", videoPath)
  hyperparameters["thresAngleBoutDetect"] = getConfig(config, "thresAngleBoutDetect", videoPath)
  # Getting parameters more specifically related to the tracking
  hyperparameters["step"] = getConfig(config, "step", videoPath)
  hyperparameters["nbList"] = getConfig(config, "nbList", videoPath)
  hyperparameters["expDecreaseFactor"] = getConfig(config, "expDecreaseFactor", videoPath)
  hyperparameters["minArea"] = getConfig(config, "minArea", videoPath)
  hyperparameters["maxArea"] = getConfig(config, "maxArea", videoPath)
  hyperparameters["minAreaBody"] = getConfig(config, "minAreaBody", videoPath)
  hyperparameters["maxAreaBody"] = getConfig(config, "maxAreaBody", videoPath)
  hyperparameters["headSize"] = getConfig(config, "headSize", videoPath)
  hyperparameters["debugExtractParams"] = getConfig(config, "debugExtractParams", videoPath)
  hyperparameters["debugTracking"]      = getConfig(config, "debugTracking", videoPath)
  hyperparameters["debugExtractBack"]   = getConfig(config, "debugExtractBack", videoPath)
  hyperparameters["debugFindWells"]     = getConfig(config, "debugFindWells", videoPath)
  hyperparameters["debugPauseBetweenTrackAndParamExtract"] = getConfig(config, "debugPauseBetweenTrackAndParamExtract", videoPath)
  hyperparameters["headEmbeded"]        = getConfig(config, "headEmbeded", videoPath)
  
  hyperparameters["halfDiameterRoiBoutDetect"] = getConfig(config, "halfDiameterRoiBoutDetect", videoPath)
  hyperparameters["thresForDetectMovementWithRawVideo"] = getConfig(config, "thresForDetectMovementWithRawVideo", videoPath)
  hyperparameters["minNbPixelForDetectMovementWithRawVideo"] = getConfig(config, "minNbPixelForDetectMovementWithRawVideo", videoPath)
  hyperparameters["frameGapComparision"]        = getConfig(config, "frameGapComparision", videoPath)
  hyperparameters["firstFrame"]                 = getConfig(config, "firstFrame", videoPath)
  hyperparameters["lastFrame"]                  = getConfig(config, "lastFrame", videoPath)
  hyperparameters["wellOutputVideoDiameter"]    = getConfig(config, "wellOutputVideoDiameter", videoPath)
  hyperparameters["fillGapFrameNb"]             = getConfig(config, "fillGapFrameNb", videoPath)
  hyperparameters["minPixelDiffForBackExtract"] = getConfig(config, "minPixelDiffForBackExtract", videoPath)
  hyperparameters["outputFolder"]               = getConfig(config, "outputFolder", videoPath)
  hyperparameters["videoName"]                  = videoName[0:(len(videoName)-4)]
  hyperparameters["detectBoutMinNbFrames"]      = getConfig(config, "detectBoutMinNbFrames", videoPath)
  hyperparameters["detectBoutMinDist"]          = getConfig(config, "detectBoutMinDist", videoPath)
  hyperparameters["nbRowsOfWells"]              = getConfig(config, "nbRowsOfWells", videoPath)
  hyperparameters["nbWellsPerRows"]             = getConfig(config, "nbWellsPerRows", videoPath)
  hyperparameters["takeTheHeadClosestToTheCenter"] = getConfig(config, "takeTheHeadClosestToTheCenter", videoPath)
  
  hyperparameters["wellsAreRectangles"]          = getConfig(config, "wellsAreRectangles", videoPath)
  hyperparameters["plotOnlyOneTailPointForVisu"] = getConfig(config, "plotOnlyOneTailPointForVisu", videoPath)
  hyperparameters["trackingPointSizeDisplay"]    = getConfig(config, "trackingPointSizeDisplay", videoPath)
  
  hyperparameters["tailAngleSmoothingFactor"]        = getConfig(config, "tailAngleSmoothingFactor", videoPath)
  hyperparameters["tailAngleMedianFilter"]           = getConfig(config, "tailAngleMedianFilter", videoPath)
  hyperparameters["minDiffBetweenSubsequentBendAmp"] = getConfig(config, "minDiffBetweenSubsequentBendAmp", videoPath)
  
  hyperparameters["centerOfMassTailTracking"]    = getConfig(config, "centerOfMassTailTracking", videoPath)
  hyperparameters["freqAlgoPosFollow"]           = getConfig(config, "freqAlgoPosFollow", videoPath)
  
  hyperparameters["centerOfMassParamStep"]         = getConfig(config, "centerOfMassParamStep", videoPath)
  hyperparameters["centerOfMassParamSegStep"]      = getConfig(config, "centerOfMassParamSegStep", videoPath)
  hyperparameters["centerOfMassParamHalfDiam"]     = getConfig(config, "centerOfMassParamHalfDiam", videoPath)
  hyperparameters["headEmbededParamGaussianBlur"] = getConfig(config, "headEmbededParamGaussianBlur", videoPath)
  
  hyperparameters["headEmbededParamInitialAngle"] = getConfig(config, "headEmbededParamInitialAngle", videoPath)
  hyperparameters["detectBoutMinAngleDiff"]        = getConfig(config, "detectBoutMinAngleDiff", videoPath)
  
  hyperparameters["coverHorizontalPortionBelowForHeadDetect"] = getConfig(config, "coverHorizontalPortionBelowForHeadDetect", videoPath)
  hyperparameters["coverHorizontalPortionAboveForHeadDetect"] = getConfig(config, "coverHorizontalPortionAboveForHeadDetect", videoPath)
  hyperparameters["coverVerticalPortionRightForHeadDetect"] = getConfig(config, "coverVerticalPortionRightForHeadDetect", videoPath)
  
  hyperparameters["windowForLocalBendMinMaxFind"] = getConfig(config, "windowForLocalBendMinMaxFind", videoPath)
  
  hyperparameters["debugValidationVideoHeading"]  = getConfig(config, "debugValidationVideoHeading", videoPath)
  hyperparameters["headingCalculationMethod"]   = getConfig(config, "headingCalculationMethod", videoPath)
  
  hyperparameters["minFirstBendValue"]   = getConfig(config, "minFirstBendValue", videoPath)
  
  hyperparameters["videoWidth"]   = getConfig(config, "videoWidth",  videoPath)
  hyperparameters["videoHeight"]  = getConfig(config, "videoHeight", videoPath)
  
  hyperparameters["debugDetectMovWithRawVideo"] = getConfig(config, "debugDetectMovWithRawVideo", videoPath)
  hyperparameters["debugDetectMovWithRawVideoShowVid"] = getConfig(config, "debugDetectMovWithRawVideoShowVid", videoPath)
  
  hyperparameters["createValidationVideo"] = getConfig(config, "createValidationVideo", videoPath)
  
  hyperparameters["minProminenceForBendsDetect"] = getConfig(config, "minProminenceForBendsDetect", videoPath)
  
  hyperparameters["debugCoverHorizontalPortionBelow"] = getConfig(config, "debugCoverHorizontalPortionBelow", videoPath)
  
  hyperparameters["minNbPeaksForBoutDetect"] = getConfig(config, "minNbPeaksForBoutDetect", videoPath)
  
  hyperparameters["freeSwimmingTailTrackingMethod"] = getConfig(config, "freeSwimmingTailTrackingMethod", videoPath)
  
  hyperparameters["findContourPrecision"] = getConfig(config, "findContourPrecision", videoPath)
  
  hyperparameters["reloadWellPositions"] = getConfig(config, "reloadWellPositions", videoPath)
  hyperparameters["reloadBackground"]    = getConfig(config, "reloadBackground", videoPath)
  hyperparameters["thresholdForBlobImg"] = getConfig(config, "thresholdForBlobImg", videoPath)
  
  hyperparameters["minTailSize"] = getConfig(config, "minTailSize", videoPath)
  hyperparameters["maxTailSize"] = getConfig(config, "maxTailSize", videoPath)
  
  hyperparameters["debugTrackingPtExtreme"] = getConfig(config, "debugTrackingPtExtreme", videoPath)
  hyperparameters["debugTrackingPtExtremeLargeVerticals"] = getConfig(config, "debugTrackingPtExtremeLargeVerticals", videoPath)
  hyperparameters["debugTrackingThreshImg"] = getConfig(config, "debugTrackingThreshImg", videoPath)

  hyperparameters["erodeSize"]  = getConfig(config, "erodeSize", videoPath)
  hyperparameters["erodeIter"]  = getConfig(config, "erodeIter", videoPath)
  hyperparameters["dilateIter"] = getConfig(config, "dilateIter", videoPath)
  
  hyperparameters["headEmbededRemoveBack"] = getConfig(config, "headEmbededRemoveBack", videoPath)
  hyperparameters["extractBackWhiteBackground"] = getConfig(config, "extractBackWhiteBackground", videoPath)
  
  hyperparameters["addBlackCircleOfHalfDiamOnHeadForBoutDetect"] = getConfig(config, "addBlackCircleOfHalfDiamOnHeadForBoutDetect", videoPath)
  
  hyperparameters["doubleCheckBendMinMaxStatus"] = getConfig(config, "doubleCheckBendMinMaxStatus", videoPath)
  hyperparameters["removeFirstSmallBend"] = getConfig(config, "removeFirstSmallBend", videoPath)
  
  hyperparameters["coverPortionForHeadDetect"] = getConfig(config, "coverPortionForHeadDetect", videoPath)
  
  hyperparameters["tailExtremityMaxJugeDecreaseCoeff"] = getConfig(config, "tailExtremityMaxJugeDecreaseCoeff", videoPath)
  
  hyperparameters["firstFrameForTracking"] = getConfig(config, "firstFrameForTracking", videoPath)
  
  hyperparameters["findHeadPositionByUserInput"] = getConfig(config, "findHeadPositionByUserInput", videoPath)
  
  hyperparameters["invertBlackWhiteOnImages"] = getConfig(config, "invertBlackWhiteOnImages", videoPath)
  
  hyperparameters["headEmbededTailTrackFindMaxDepthInitialMaxDepth"] = getConfig(config, "headEmbededTailTrackFindMaxDepthInitialMaxDepth", videoPath)
  
  hyperparameters["debugHeadEmbededFindNextPoints"] = getConfig(config, "debugHeadEmbededFindNextPoints", videoPath)
  
  hyperparameters["headEmbededTeresaNicolson"] = getConfig(config, "headEmbededTeresaNicolson", videoPath)
  
  hyperparameters["boutEdgesWhereZeros"] = getConfig(config, "boutEdgesWhereZeros", videoPath)
  
  hyperparameters["noBoutsDetection"]    = getConfig(config, "noBoutsDetection", videoPath)
  
  hyperparameters["setPixDiffBoutDetectParameters"] = getConfig(config, "setPixDiffBoutDetectParameters", videoPath)
  
  hyperparameters["midlineIsInBlobTrackingOptimization"] = getConfig(config, "midlineIsInBlobTrackingOptimization", videoPath)
  
  hyperparameters["debugHeadingCalculation"] = getConfig(config, "debugHeadingCalculation", videoPath)
  
  hyperparameters["accentuateFrameForManualTailExtremityFind"] = getConfig(config, "accentuateFrameForManualTailExtremityFind", videoPath)
  
  hyperparameters["popUpAlgoFollow"] = getConfig(config, "popUpAlgoFollow", videoPath)
  
  hyperparameters["paramGaussianBlur"] = getConfig(config, "paramGaussianBlur", videoPath)
  
  hyperparameters["generateAllTimeTailAngleGraph"] = getConfig(config, "generateAllTimeTailAngleGraph", videoPath)
  
  hyperparameters["generateAllTimeTailAngleGraphLineWidth"] = getConfig(config, "generateAllTimeTailAngleGraphLineWidth", videoPath)
  
  hyperparameters["extractAdvanceZebraParameters"] = getConfig(config, "extractAdvanceZebraParameters", videoPath)
  
  hyperparameters["noWellDetection"] = getConfig(config, "noWellDetection", videoPath)
  
  hyperparameters["nbAnimalsPerWell"] = getConfig(config, "nbAnimalsPerWell", videoPath)
  
  hyperparameters["oneWellManuallyChosenTopLeft"] = getConfig(config, "oneWellManuallyChosenTopLeft", videoPath)
  hyperparameters["oneWellManuallyChosenBottomRight"] = getConfig(config, "oneWellManuallyChosenBottomRight", videoPath)
  
  hyperparameters["noChecksForBoutSelectionInExtractParams"] = getConfig(config, "noChecksForBoutSelectionInExtractParams", videoPath)
  
  hyperparameters["validationVideoPlotHeading"] = getConfig(config, "validationVideoPlotHeading", videoPath)
  
  hyperparameters["multipleHeadTrackingIterativelyRelaxAreaCriteria"] = getConfig(config, "multipleHeadTrackingIterativelyRelaxAreaCriteria", videoPath)
  
  hyperparameters["closePopUpWindowAtTheEnd"] = getConfig(config, "closePopUpWindowAtTheEnd", videoPath)
  
  hyperparameters["multipleAnimalTrackingAdvanceAlgorithm"] = getConfig(config, "multipleAnimalTrackingAdvanceAlgorithm", videoPath)
  
  hyperparameters["multipleAnimalTrackingAdvanceAlgorithmMarginX"] = getConfig(config, "multipleAnimalTrackingAdvanceAlgorithmMarginX", videoPath)
  hyperparameters["multipleAnimalTrackingAdvanceAlgorithmDist1"] = getConfig(config, "multipleAnimalTrackingAdvanceAlgorithmDist1", videoPath)
  hyperparameters["multipleAnimalTrackingAdvanceAlgorithmDist2"] = getConfig(config, "multipleAnimalTrackingAdvanceAlgorithmDist2", videoPath)
  
  hyperparameters["perBoutOutput"] = getConfig(config, "perBoutOutput", videoPath)
  hyperparameters["perBoutOutputVideoStartStopFrameMargin"] = getConfig(config, "perBoutOutputVideoStartStopFrameMargin", videoPath)
  hyperparameters["perBoutOutputYaxis"] = getConfig(config, "perBoutOutputYaxis", videoPath)
  
  hyperparameters["smoothTailHeadEmbeded"] = getConfig(config, "smoothTailHeadEmbeded", videoPath)
  hyperparameters["nbTailPoints"] = getConfig(config, "nbTailPoints", videoPath)
  
  hyperparameters["curvatureMedianFilterSmoothingWindow"] = getConfig(config, "curvatureMedianFilterSmoothingWindow", videoPath)
  
  hyperparameters["headEmbededParamTailDescentPixThreshStop"] = getConfig(config, "headEmbededParamTailDescentPixThreshStop", videoPath)
  
  hyperparameters["automaticallySetSomeOfTheHeadEmbededHyperparameters"] = getConfig(config, "automaticallySetSomeOfTheHeadEmbededHyperparameters", videoPath)
  hyperparameters["headEmbededAutoSet_ExtendedDescentSearchOption"] = getConfig(config, "headEmbededAutoSet_ExtendedDescentSearchOption", videoPath)
  hyperparameters["headEmbededAutoSet_BackgroundExtractionOption"] = getConfig(config, "headEmbededAutoSet_BackgroundExtractionOption", videoPath)
  
  hyperparameters["trackTail"] = getConfig(config, "trackTail", videoPath)
  
  hyperparameters["adjustDetectMovWithRawVideo"] = getConfig(config, "adjustDetectMovWithRawVideo", videoPath)
  
  hyperparameters["adjustHeadEmbededTracking"] = getConfig(config, "adjustHeadEmbededTracking", videoPath)
  hyperparameters["adjustFreelySwimTracking"] = getConfig(config, "adjustFreelySwimTracking", videoPath)
  
  hyperparameters["overwriteFirstStepValue"] = getConfig(config, "overwriteFirstStepValue", videoPath)
  hyperparameters["overwriteLastStepValue"]  = getConfig(config, "overwriteLastStepValue", videoPath)
  hyperparameters["overwriteNbOfStepValues"] = getConfig(config, "overwriteNbOfStepValues", videoPath)
  
  hyperparameters["headEmbededParamTailDescentPixThreshStopOverwrite"] = getConfig(config, "headEmbededParamTailDescentPixThreshStopOverwrite", videoPath)
  
  hyperparameters["exitAfterBackgroundExtraction"] = getConfig(config, "exitAfterBackgroundExtraction", videoPath)
  
  hyperparameters["nbImagesForBackgroundCalculation"] = getConfig(config, "nbImagesForBackgroundCalculation", videoPath)
  
  hyperparameters["authorizedRelativeLengthTailEnd"] = getConfig(config, "authorizedRelativeLengthTailEnd", videoPath)
  
  hyperparameters["initialTailPortionMaxSegmentDiffAngleCutOffPos"] = getConfig(config, "initialTailPortionMaxSegmentDiffAngleCutOffPos", videoPath)
  hyperparameters["initialTailPortionMaxSegmentDiffAngleValue"] = getConfig(config, "initialTailPortionMaxSegmentDiffAngleValue", videoPath)
  
  hyperparameters["windowForBoutDetectWithAngle"] = getConfig(config, "windowForBoutDetectWithAngle", videoPath)
  
  hyperparameters["forceBlobMethodForHeadTracking"] = getConfig(config, "forceBlobMethodForHeadTracking", videoPath)
  
  hyperparameters["headEmbededRetrackIfWeirdInitialTracking"] = getConfig(config, "headEmbededRetrackIfWeirdInitialTracking", videoPath)
  
  hyperparameters["overwriteHeadEmbededParamGaussianBlur"] = getConfig(config, "overwriteHeadEmbededParamGaussianBlur", videoPath)
  
  hyperparameters["calculateAllTailAngles"] = getConfig(config, "calculateAllTailAngles", videoPath)
  
  hyperparameters["nbPointsToIgnoreAtCurvatureBeginning"] = getConfig(config, "nbPointsToIgnoreAtCurvatureBeginning", videoPath)
  
  hyperparameters["nbPointsToIgnoreAtCurvatureEnd"] = getConfig(config, "nbPointsToIgnoreAtCurvatureEnd", videoPath)
  
  hyperparameters["imagePreProcessMethod"] = getConfig(config, "imagePreProcessMethod", videoPath)
  hyperparameters["imagePreProcessParameters"] = getConfig(config, "imagePreProcessParameters", videoPath)
  
  hyperparameters["backgroundPreProcessMethod"] = getConfig(config, "backgroundPreProcessMethod", videoPath)
  hyperparameters["backgroundPreProcessParameters"] = getConfig(config, "backgroundPreProcessParameters", videoPath)
  
  hyperparameters["backgroundExtractionWithOnlyTwoFrames"] = getConfig(config, "backgroundExtractionWithOnlyTwoFrames", videoPath)
  
  hyperparameters["checkThatMovementOccurInVideo"] = getConfig(config, "checkThatMovementOccurInVideo", videoPath)
  hyperparameters["checkThatMovementOccurInVideoMedianFilterWindow"] = getConfig(config, "checkThatMovementOccurInVideoMedianFilterWindow", videoPath)
  
  hyperparameters["findRectangleWellArea"] = getConfig(config, "findRectangleWellArea", videoPath)
  hyperparameters["rectangleWellAreaImageThreshold"] = getConfig(config, "rectangleWellAreaImageThreshold", videoPath)
  hyperparameters["rectangleWellErodeDilateKernelSize"] = getConfig(config, "rectangleWellErodeDilateKernelSize", videoPath)
  hyperparameters["rectangularWellsInvertBlackWhite"] = getConfig(config, "rectangularWellsInvertBlackWhite", videoPath)
  hyperparameters["rectangularWellStretchPercentage"] = getConfig(config, "rectangularWellStretchPercentage", videoPath)
  hyperparameters["rectangleWellAreaTolerancePercentage"] = getConfig(config, "rectangleWellAreaTolerancePercentage", videoPath)

  hyperparameters["adjustRectangularWellsDetect"] = getConfig(config, "adjustRectangularWellsDetect", videoPath)
  
  hyperparameters["outputValidationVideoFps"] = getConfig(config, "outputValidationVideoFps", videoPath)
  
  hyperparameters["checkAllContourForTailExtremityDetect"] = getConfig(config, "checkAllContourForTailExtremityDetect", videoPath)
  hyperparameters["considerHighPointForTailExtremityDetect"] = getConfig(config, "considerHighPointForTailExtremityDetect", videoPath)
  
  hyperparameters["debugEyeTracking"]                   = getConfig(config, "debugEyeTracking", videoPath)
  hyperparameters["debugEyeTrackingAdvanced"]           = getConfig(config, "debugEyeTrackingAdvanced", videoPath)
  hyperparameters["eyeTracking"]                        = getConfig(config, "eyeTracking", videoPath)
  hyperparameters["headCenterToMidEyesPointDistance"]   = getConfig(config, "headCenterToMidEyesPointDistance", videoPath)
  hyperparameters["eyeBinaryThreshold"]                 = getConfig(config, "eyeBinaryThreshold", videoPath)
  hyperparameters["midEyesPointToEyeCenterMaxDistance"] = getConfig(config, "midEyesPointToEyeCenterMaxDistance", videoPath)
  hyperparameters["eyeHeadingSearchAreaHalfDiameter"]   = getConfig(config, "eyeHeadingSearchAreaHalfDiameter", videoPath)
  hyperparameters["headingLineValidationPlotLength"]    = getConfig(config, "headingLineValidationPlotLength", videoPath)
  hyperparameters["saveSuperStructToMatlab"]            = getConfig(config, "saveSuperStructToMatlab", videoPath)
  hyperparameters["validationVideoPlotAnimalNumber"]    = getConfig(config, "validationVideoPlotAnimalNumber", videoPath)
  hyperparameters["postProcessMultipleTrajectories"]    = getConfig(config, "postProcessMultipleTrajectories", videoPath)
  hyperparameters["postProcessMaxDistanceAuthorized"]   = getConfig(config, "postProcessMaxDistanceAuthorized", videoPath)
  hyperparameters["postProcessMaxDisapearanceFrames"]   = getConfig(config, "postProcessMaxDisapearanceFrames", videoPath)
  hyperparameters["computeEyesHeadingPlot"]             = getConfig(config, "computeEyesHeadingPlot", videoPath)
  hyperparameters["adjustMinPixelDiffForBackExtract_nbBlackPixelsMax"] = getConfig(config, "adjustMinPixelDiffForBackExtract_nbBlackPixelsMax", videoPath)
  
  hyperparameters["fixedHeadPositionX"] = getConfig(config, "fixedHeadPositionX", videoPath)
  hyperparameters["fixedHeadPositionY"] = getConfig(config, "fixedHeadPositionY", videoPath)
  
  hyperparameters["recalculateForegroundImageBasedOnBodyArea"] = getConfig(config, "recalculateForegroundImageBasedOnBodyArea", videoPath)
  
  hyperparameters["detectMouthInsteadOfHeadTwoSides"] = getConfig(config, "detectMouthInsteadOfHeadTwoSides", videoPath)
  hyperparameters["findCenterOfAnimalByIterativelyDilating"] = getConfig(config, "findCenterOfAnimalByIterativelyDilating", videoPath)
  
  hyperparameters["backgroundExtractionForceUseAllVideoFrames"] = getConfig(config, "backgroundExtractionForceUseAllVideoFrames", videoPath)
  
  hyperparameters["readjustCenterOfMassIfNotInsideContour"] = getConfig(config, "readjustCenterOfMassIfNotInsideContour", videoPath)
  
  if len(argv) > 5 and not(argv[0] == "getTailExtremityFirstFrame"):
    i = 5
    while i < len(argv):
      print("command line hyperparameter change:", argv[i], argv[i+1])
      
      if not(argv[i] in ["debugPauseBetweenTrackAndParamExtract", "outputFolder", "coverPortionForHeadDetect", "freeSwimmingTailTrackingMethod", "findContourPrecision", "headingCalculationMethod"]):
        try:
          hyperparameters[argv[i]] = int(argv[i+1])
        except:
          hyperparameters[argv[i]] = float(argv[i+1])
        
      else:
        hyperparameters[argv[i]] = argv[i+1]
      
      i = i + 2
  
  
  if hyperparameters["setPixDiffBoutDetectParameters"] > 0:
    hyperparameters["debugPauseBetweenTrackAndParamExtract"] = "justExtractParamFromPreviousTrackData"
    hyperparameters["createValidationVideo"]                 = 1 
    hyperparameters["reloadBackground"]                      = 1 
    hyperparameters["debugDetectMovWithRawVideo"]            = 1
    hyperparameters["debugDetectMovWithRawVideoShowVid"]     = 1
    hyperparameters["reloadWellPositions"]                   = 1
    
  if hyperparameters["setPixDiffBoutDetectParameters"] == 2:
    hyperparameters["debugDetectMovWithRawVideoShowVid"]     = 0
  
  # if globalVariables["limitedVersion"] == 1:
    # limitNbFrames = 1000
    # if hyperparameters["lastFrame"] - hyperparameters["firstFrame"] > limitNbFrames:
      # hyperparameters["lastFrame"] = int(hyperparameters["firstFrame"]) + limitNbFrames
  
  return hyperparameters
  
  
def getHyperparametersSimple(configTemp):
  hyperparameters = configDefault.copy()
  for index in configTemp:
    hyperparameters[index] = configTemp[index]
  return hyperparameters

